import { useMemo, ReactElement, useState } from 'react';
import { NodeAttribute, QueryGraphEdges, SchemaReactflowEntityNode, handleDataFromReactflowToDataId, toHandleId } from '../../model';
import { Handle, Position, useUpdateNodeInternals } from 'reactflow';
import { Abc, CalendarToday, Map, Numbers, Place, QuestionMarkOutlined } from '@mui/icons-material';
import { Icon } from '@graphpolaris/shared/lib/components/icon';
import { PillHandle } from '@graphpolaris/shared/lib/components/pills/PillHandle';
import { pillDropdownPadding } from '@graphpolaris/shared/lib/components/pills/pill.const';
import { Button, TextInput, useAppDispatch, useQuerybuilderAttributesShown } from '../../..';
import { attributeShownToggle } from '@graphpolaris/shared/lib/data-access/store/querybuilderSlice';
import { isEqual } from 'lodash-es';

type PillDropdownProps = {
  node: SchemaReactflowEntityNode;
  attributes: NodeAttribute[];
  attributeEdges: (QueryGraphEdges | undefined)[];
  hovered: boolean;
  handleBeingDraggedIdx: number;
  onHandleMouseDown: (attribute: NodeAttribute, i: number, event: React.MouseEvent) => void;
  onMouseEnterDropdown?: (event: React.MouseEvent) => void;
  onMouseLeaveDropdown?: (event: React.MouseEvent) => void;
};

type IconMapType = {
  [key: string]: ReactElement;
};

const IconMap: IconMapType = {
  temporal: <CalendarToday />,
  spatial: <Map />,
  numerical: <Numbers />,
  categorical: <Abc />,
  default: <Place />,
};

export const PillDropdown = (props: PillDropdownProps) => {
  const forceOpen = false;
  const [filter, setFilter] = useState<string>('');
  const dispatch = useAppDispatch();
  const attributesBeingShown = useQuerybuilderAttributesShown();

  const attributesOfInterest = useMemo(() => {
    return props.attributes.map((attribute) =>
      attributesBeingShown.findIndex((x) => isEqual(x, attribute.handleData)) === -1 ? false : true,
    );
  }, [attributesBeingShown]);

  return (
    <div
      className={'border-[1px] border-secondary-200 divide-y divide-secondary-200 !z-50'}
      onMouseEnter={(e) => {
        if (props.onMouseEnterDropdown) props.onMouseEnterDropdown(e);
      }}
      onMouseLeave={(e) => {
        if (props.onMouseLeaveDropdown) props.onMouseLeaveDropdown(e);
      }}
    >
      {attributesOfInterest &&
        attributesOfInterest.map((showing, i) => {
          if (showing === false) return null;

          const attribute = props.attributes[i];
          if (attribute.handleData.attributeName === undefined) {
            throw new Error('attribute.handleData.attributeName is undefined');
          }

          return (
            <div
              className="px-2 py-1 bg-secondary-100 flex justify-between items-center"
              key={(attribute.handleData.attributeName || '') + i}
              onMouseDown={(event: React.MouseEvent) => {
                props.onHandleMouseDown(attribute, i, event);
              }}
            >
              <p className="truncate text-[0.6rem]">{attribute.handleData.attributeName}</p>
              {attribute.handleData?.attributeDimension && <Icon component={IconMap[attribute.handleData.attributeDimension]} size={16} />}
              <PillHandle
                mr={-pillDropdownPadding}
                handleTop="auto"
                position={Position.Right}
                className={'fill-accent-500 stroke-white'}
                type="square"
              >
                <Handle
                  id={toHandleId(handleDataFromReactflowToDataId(props.node, attribute))}
                  type="source"
                  position={Position.Right}
                  className={'!rounded-none !bg-transparent !w-full !h-full !right-0 !left-0 !border-0'}
                ></Handle>
              </PillHandle>
            </div>
          );
        })}
      {(props.hovered || forceOpen) && (
        <>
          <h4 className="p-1 bg-white border-t-[2px] font-semibold text-2xs">Available Attributes:</h4>
          <TextInput type={'text'} placeholder="Filter" size="xs" className="!p-0.5" value={filter} onChange={(v) => setFilter(v)} />
          <div className="max-h-28 overflow-auto flex flex-col bg-white">
            {props.attributes.map((attribute, i) => {
              if (filter && !attribute.handleData.attributeName?.toLowerCase().includes(filter.toLowerCase())) return null;
              if (attribute.handleData.attributeName === undefined) {
                throw new Error('attribute.handleData.attributeName is undefined');
              }

              return (
                <Button
                  key={(attribute.handleData.attributeName || '') + i}
                  iconComponent={attribute?.handleData?.attributeDimension ? IconMap[attribute.handleData.attributeDimension] : undefined}
                  iconPosition="trailing"
                  className={`w-full ${attributesOfInterest[i] ? 'bg-secondary-100' : 'bg-white'} justify-between rounded-none text-[0.7em]`}
                  variant="ghost"
                  size={'xs'}
                  label={attribute.handleData.attributeName}
                  onClick={(event: React.MouseEvent) => {
                    dispatch(attributeShownToggle(attribute.handleData));
                  }}
                ></Button>
              );
            })}
          </div>
        </>
      )}
    </div>
  );
};
